type 'a table = 'a Matrix.t

type 'a sudoku = int * 'a table
type problem = int option sudoku
type solution = int sudoku

type guess = Fixed of int | Choices of int list
type guesses = guess sudoku

exception InternalError

let string_of_int_list (l: int list): string =
  "{" ^
  (String.concat ", " (List.map string_of_int l)) ^
  "}"

let getSize
    ((s, _): 'a sudoku)
    : int =
  s

let getElem
    ((_, m): 'a sudoku)
    ((r, c): int * int)
    : 'a =
  Matrix.get m (r-1, c-1)

let setElem
    ((_, m): 'a sudoku)
    ((r, c): int * int)
    (a: 'a)
    : unit =
  Matrix.set m (r-1, c-1) a

let printSudoku
    (sudoku: 'a sudoku)
    (printer: 'a -> unit)
    : unit =
  let size = getSize sudoku
  in
  for r = 1 to size*size do
    for c = 1 to size*size do
      printer (getElem sudoku (r, c));
      if c <> size*size then
      begin
        if c mod size = 0 && c > 0 then
          Printf.printf " ";
        Printf.printf " ";
      end;
    done;
    if r mod size = 0 && r > 0 && r != size*size then
      Printf.printf "\n";
    Printf.printf "\n";
  done

let inc (x: int ref): unit =
  x := !x + 1

let range (min: int) (max: int): int list =
  let rec range'
      (min: int)
      (max: int)
      (l: int list)
      : int list =
    if min > max then
      List.rev l
    else
      range' (min+1) (max) (min::l)
  in
  range' min max []

let create 
    (s: int)
    (v: int option list)
    : problem =
  let sudoku = (s, Matrix.create (s*s, s*s) None)
  in
  let index = ref 0
  in
  for r = 1 to s*s do
    for c = 1 to s*s do
      setElem sudoku (r, c) (List.nth v !index);
      inc index;
    done;
  done;
  sudoku

let solve
    (p: problem)
    : solution option =
  let s = getSize p
  in
  let s2 = s * s
  in
  let any = range 1 s2
  in

  let printGuess (e: guess): unit =
    match e with
    | Fixed v   ->
        Printf.printf "%d" v
    | Choices l ->
        Printf.printf "%s" (string_of_int_list l)
  in

  let printPartialSolution (g: guesses): unit =
    let printer (e: guess): unit =
      match e with
      | Fixed v -> Printf.printf "%d" v
      | _       -> Printf.printf "-"
    in
    Printf.printf "Partial Solution:\n";
    printSudoku g printer;
    Printf.printf "\n\n";
  in

  let printGuesses (g: guesses): unit =
    let size = getSize g
    in
    for r = 1 to size*size do
      for c = 1 to size*size do
        Printf.printf "%d:%d: " r c;
        printGuess (getElem g (r, c));
        Printf.printf "\n";
      done;
      Printf.printf "\n";
    done;
    Printf.printf "\n";
    Printf.printf "\n";
  in

  let initialGuess (p: problem): guesses =
    let g = (s, Matrix.create (s*s, s*s) (Choices any))
    in
    for r = 1 to s2 do
      for c = 1 to s2 do
        match getElem p (r, c) with
        | Some v ->
            setElem g (r, c) (Fixed v)
        | None   ->
            ()
      done;
    done;
    g
  in

  let updateGuesses (g: guesses): bool =
    let updated = ref false
    in

    let updateRow (r: int): unit =
      let fixed =
        List.fold_left
          (fun l c ->
            match getElem g (r, c) with
            | Fixed v -> v::l
            | _       -> l
          )
          []
          any
      in
      let remaining = ListSet.diff any fixed
      in
      for c = 1 to s2 do
        match getElem g (r, c) with
        | Fixed _   -> ()
        | Choices l ->
            let l' = ListSet.intersection l remaining
            in
            if l <> l' then
            begin
              updated := true;
              setElem g (r, c) (Choices l');
            end;
      done;
    in

    let updateRows (): unit =
      List.iter updateRow any
    in

    let updateCol (c: int): unit =
      let fixed =
        List.fold_left
          (fun l r ->
            match getElem g (r, c) with
            | Fixed v -> v::l
            | _       -> l
          )
          []
          any
      in
      let remaining = ListSet.diff any fixed
      in
      for r = 1 to s2 do
        match getElem g (r, c) with
        | Fixed _   -> ()
        | Choices l ->
            let l' = ListSet.intersection l remaining
            in
            if l <> l' then
            begin
              updated := true;
              setElem g (r, c) (Choices l');
            end;
      done;
    in

    let updateCols (): unit =
      List.iter updateCol any
    in

    let updateBlock ((br, bc): int * int): unit =
      let bbr = (br-1) * s
      in
      let bbc = (bc-1) * s
      in
      let fixed =
        List.fold_left
            (fun l r ->
              List.fold_left
                  (fun l c ->
                    match getElem g (bbr+r, bbc+c) with
                    | Fixed v -> v::l
                    | _       -> l
                  )
                  l
                  (range 1 s)
            )
            []
            (range 1 s)
      in
      let remaining = ListSet.diff any fixed
      in
      for r = 1 to s do
        for c = 1 to s do
          match getElem g (bbr+r, bbc+c) with
          | Fixed _   -> ()
          | Choices l ->
              let l' = ListSet.intersection l remaining
              in
              if l <> l' then
              begin
                updated := true;
                setElem g (bbr+r, bbc+c) (Choices l');
              end;
        done;
      done
    in

    let updateBlocks (): unit =
      for br = 1 to s do
        for bc = 1 to s do
          updateBlock (br, bc)
        done;
      done
    in

    updateRows ();
    updateCols ();
    updateBlocks ();
    !updated
  in

  let simplifyGuesses (g: guesses): unit =
    for r = 1 to s2 do
      for c = 1 to s2 do
        match getElem g (r, c) with
        | Choices [v] -> setElem g (r, c) (Fixed v)
        | _           -> ()
      done;
    done
  in

  let isSolved (g: guesses): bool =
    let solved = ref true
    in
    for r = 1 to s2 do
      for c = 1 to s2 do
        match getElem g (r, c) with
        | Choices _ -> solved := false
        | _         -> ()
      done;
    done;
    !solved
  in

  let getSolution (g: guesses): solution =
    let solution = (s, Matrix.create (s*s, s*s) 0)
    in
    for r = 1 to s2 do
      for c = 1 to s2 do
        match getElem g (r, c) with
        | Fixed v -> setElem solution (r, c) v
        | _       -> raise InternalError
      done;
    done;
    solution
  in

  let g = initialGuess p
  in

  printPartialSolution g;

  while updateGuesses g do
    simplifyGuesses g;
    printPartialSolution g;
  done;

  if isSolved g then
    Some (getSolution g)
  else
    None

let printProblem (p: problem): unit =
  let printer (e: int option): unit =
    match e with
    | Some v -> Printf.printf "%d" v
    | None   -> Printf.printf "-"
  in
  printSudoku p printer

let printSolution (s: solution): unit =
  let printer (e: int): unit =
    Printf.printf "%d" e
  in
  printSudoku s printer
